"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2021 Marco Esposito, Leonardo Picchiami.
"""

import COPASI
from apricopt.solving.whitebox.WhiteBoxSolver import WhiteBoxSolver
from apricopt.model.Model import Model
from apricopt.solving.whitebox.COPASI.COPASIOptimisationMethod import COPASIOptimisationMethod
from apricopt.simulation.COPASI.COPASIModelInstance import COPASIModelInstance
from apricopt.simulation.SimulationEngine import SimulationEngine
from apricopt.sampling.Sampler import Sampler
from typing import Dict, Union, Tuple, List
import sys


def copasi_solver_log(data_model: COPASI.CDataModel, opt_task: COPASI.COptTask):
    opt_problem = opt_task.getProblem()
    standard_reports = data_model.getReportDefinitionList()
    report = standard_reports.createReportDefinition("Apricopt Report", "Output for optimization")
    # std = standard_reports.get(3)
    report.setTaskType(COPASI.CTaskEnum.Task_optimization)
    report.setIsTable(False)
    report.setSeparator(COPASI.CCopasiReportSeparator("\t\t\t"))

    header = report.getHeaderAddr()
    body = report.getBodyAddr()
    footer = report.getFooterAddr()

    header.push_back(
        COPASI.CRegisteredCommonName(COPASI.CDataString("======= APRICOPT =========\n\n").getCN().getString()))
    header.push_back(
        COPASI.CRegisteredCommonName(
            COPASI.CDataString("=== Optimization through COPASI solver ===\n\n\n").getCN().getString()))
    header.push_back(COPASI.CRegisteredCommonName(COPASI.CDataString(f"Solver settings: \n\n").getCN().getString()))
    header.push_back(
        COPASI.CRegisteredCommonName(
            COPASI.CDataString(f"{opt_task.getMethod().printToString()}\n\n\n").getCN().getString()))
    header.push_back(COPASI.CRegisteredCommonName(COPASI.CDataString(f"Optimization Log:\n\n").getCN().getString()))

    header.push_back(COPASI.CRegisteredCommonName(COPASI.CDataString("Function Evaluation").getCN().getString()))
    header.push_back(COPASI.CRegisteredCommonName(COPASI.CDataString("\t").getCN().getString()))
    header.push_back(COPASI.CRegisteredCommonName(COPASI.CDataString("Objective Value").getCN().getString()))
    header.push_back(COPASI.CRegisteredCommonName(COPASI.CDataString("\t\t").getCN().getString()))
    header.push_back(COPASI.CRegisteredCommonName(COPASI.CDataString("Optimization Parameters").getCN().getString()))

    body.push_back(
        COPASI.CRegisteredCommonName(opt_problem.getObject(
            COPASI.CCommonName("Reference=Function Evaluations")).getCN().getString()))
    body.push_back(COPASI.CRegisteredCommonName(report.getSeparator().getCN().getString()))

    body.push_back(COPASI.CRegisteredCommonName(opt_problem.getObject(
        COPASI.CCommonName("Reference=Best Value")).getCN().getString()))
    body.push_back(COPASI.CRegisteredCommonName(COPASI.CDataString("\t\t\t").getCN().getString()))

    body.push_back(
        COPASI.CRegisteredCommonName(opt_problem.getObject(
            COPASI.CCommonName("Reference=Best Parameters")).getCN().getString()))

    footer.push_back(COPASI.CRegisteredCommonName(COPASI.CDataString("\n\n\n").getCN().getString()))
    footer.push_back(COPASI.CRegisteredCommonName(opt_task.getObject(
        COPASI.CCommonName("Object=Result")).getCN().getString()))

    return report


def get_optimisation_item_from_sbml(model: Model, item_id: str) -> \
        Union[Union[COPASI.CModelValue, COPASI.CMetab], None]:
    data_model: COPASI.CDataModel = model.instance.model_obj
    if isinstance(model.instance, COPASIModelInstance):
        sbml_obj = model.instance.getModelValueBySBMLIdIfPresent(data_model.getModel(), item_id)
        if sbml_obj:
            return sbml_obj
        sbml_obj = model.instance.getMetaboliteBySBMLIdIfPresent(data_model.getModel(), item_id)
        if not sbml_obj:
            raise ValueError("The Objective function or constraint is not encoded in the model")
    else:
        raise ValueError("It is not possibile to use a Simulation Engine different from COPASI")
    return sbml_obj


class COPASISolver(WhiteBoxSolver):
    MINIMISE = 0
    MAXIMISE = 1

    def __init__(self, optimisation_method: COPASIOptimisationMethod, log_filename: str):
        super(COPASISolver, self).__init__()
        self.optimisation_method = optimisation_method
        self.log_filename = log_filename
        self.n_restarts = 0
        self.seed = None
        self.restart_treatments: List[Dict[str, float]] = []
        self.restart_log_folder = None
        self.minimize_maximize = COPASISolver.MINIMISE

    def build_copasi_optimisation_problem(self, model: Model, opt_problem: COPASI.COptProblem) -> None:
        if not isinstance(model.instance, COPASIModelInstance):
            raise TypeError("The instance field of the model must contain an object of type COPASIModelInstance.")

        if self.minimize_maximize == COPASISolver.MAXIMISE:
            opt_problem.setMaximize(True)

        data_model: COPASI.CDataModel = model.instance.model_obj
        objective_obj = get_optimisation_item_from_sbml(model, model.objective.id)
        objective_function = f'<{objective_obj.getObject(COPASI.CCommonName("Reference=Value")).getCN().getString()}>'
        opt_problem.setObjectiveFunction(objective_function)

        copasi_optimisation_constraints = opt_problem.getParameter("OptimizationConstraintList")
        for constr in model.constraints:
            constr_sbml = get_optimisation_item_from_sbml(model, constr.id)
            item = COPASI.COptItem(data_model)
            item.setObjectCN(COPASI.CCommonName(constr_sbml.getObject(COPASI.CCommonName("Reference=Value")).getCN()))
            item.setUpperBound(COPASI.CCommonName(f"{constr.upper_bound}"))
            item.setLowerBound(COPASI.CCommonName(f"{constr.lower_bound}"))
            copasi_optimisation_constraints.addParameter(item)

        for param_id, param in sorted(model.parameters.items(), key=lambda i: i[1].id):
            param_sbml = get_optimisation_item_from_sbml(model, param.id)
            item = opt_problem.addOptItem(
                COPASI.CCommonName(param_sbml.getObject(COPASI.CCommonName("Reference=InitialValue")).getCN())
            )
            item.setUpperBound(COPASI.CCommonName(f"{param.upper_bound}"))
            item.setLowerBound(COPASI.CCommonName(f"{param.lower_bound}"))
            item.setStartValue(param.nominal_value)

    def _change_starting_point(self, opt_problem: COPASI.COptProblem, treatment: Dict[str, float]) -> None:
        opt_item_list = opt_problem.getParameter('OptimizationItemList')
        param_list = [param for param in treatment.keys()]
        param_list.sort()
        for i in range(opt_item_list.size()):
            item = opt_item_list.getParameter(i)
            item.setStartValue(treatment[param_list[i]])

    def _simulation_config_for_sub_task(self, model: Model, horizon: float) -> None:
        data_model: COPASI.CDataModel = model.instance.model_obj
        simulation_task = data_model.getTask("Time-Course")
        assert simulation_task is not None
        copasi_model = data_model.getModel()

        copasi_model.setInitialTime(0.0)
        simulation_task.setMethodType(COPASI.CTaskEnum.Method_deterministic)
        simulation_task.setScheduled(True)

        model.instance.set_simulation_duration_with_step_size(horizon, model.time_step)

    def _run_and_get_output(self, model: Model, opt_task: COPASI.COptTask, opt_problem: COPASI.COptProblem) \
            -> Tuple[Dict[str, float], float, int, Dict[str, Union[float, list]]]:
        result = False
        try:
            result = opt_task.process(True)
        except:
            sys.stderr.write("Running the optimization failed.\n")
            exit(2)
        if not result:
            sys.stderr.write("Running the optimization failed.\n")
            exit(2)

        params_list = [param_id.id for param_id in model.parameters.values()]
        params_list.sort()
        solution_variables = opt_problem.getSolutionVariables()
        optimized_parameters: Dict[str, float] = dict()
        for i in range(len(params_list)):
            optimized_parameters[params_list[i]] = solution_variables.get(i)

        objective_value = opt_problem.getSolutionValue()
        function_evaluations = opt_problem.getFunctionEvaluations()
        statistics: Dict[str, Union[float, list]] = {
            'failed_func_evaluations_nan': opt_problem.getFailedEvaluationsNaN(),
            'failed_func_evaluations_exception': opt_problem.getFailedEvaluationsExc(),
            'execution_time': opt_problem.getExecutionTime(),
        }
        return optimized_parameters, objective_value, function_evaluations, statistics

    def _change_log_path_for_restart(self, opt_task, index: int, log_report):
        opt_task.getReport().setReportDefinition(log_report)
        opt_task.getReport().setTarget(f"{self.restart_log_folder}/restart_{index}_log.txt")
        opt_task.getReport().setAppend(False)

    def set_objective_maximization(self):
        self.minimize_maximize = COPASISolver.MAXIMISE

    def set_restart_configuration(self, seed: int, n_restarts: int,
                                  restart_treatments: List[Dict[str, float]], restart_log_folder: str) -> None:
        self.seed = seed
        self.n_restarts = n_restarts
        self.restart_treatments = restart_treatments
        self.restart_log_folder = restart_log_folder

    def solve(self, model: Model, H: float, sim_engine: SimulationEngine,
              solver_params: list) -> (Dict[str, float], float, dict):
        # Set Time-Course as subtask
        self._simulation_config_for_sub_task(model, H)

        # Set optimization task and time-course subtask type
        data_model: COPASI.CDataModel = model.instance.model_obj
        opt_task: COPASI.COptTask = data_model.getTask('Optimization')
        assert opt_task is not None
        self.optimisation_method.set_parameters_configuration(opt_task, solver_params)
        opt_problem: COPASI.COptProblem = opt_task.getProblem()
        assert opt_problem is not None
        opt_problem.setSubtaskType(COPASI.CTaskEnum.Task_timeCourse)
        self.build_copasi_optimisation_problem(model, opt_problem)

        log_report = copasi_solver_log(data_model, opt_task)
        opt_task.getReport().setReportDefinition(log_report)
        opt_task.getReport().setTarget(self.log_filename)
        opt_task.getReport().setAppend(False)

        optimized_parameters, objective_value, \
            function_evaluations, statistics = self._run_and_get_output(model, opt_task, opt_problem)

        if self.n_restarts > 0:
            statistics['restarts_info'] = []
            sampler = Sampler(model, sim_engine, H, self.seed)
            for i in range(self.n_restarts):
                print(f"Restart {i+1}")
                if i < len(self.restart_treatments):
                    self._change_starting_point(opt_problem, self.restart_treatments[i])
                else:
                    sampled_starting_point = sampler.random_sampling_admissible_parameter_with_respect_constraints()
                    self._change_starting_point(opt_problem, sampled_starting_point)
                self._change_log_path_for_restart(opt_task, i+1, log_report)
                print(f"\t\tStart optimisation restart {i+1}..")
                result = self._run_and_get_output(model, opt_task, opt_problem)
                statistics['restarts_info'].append(result)

        return optimized_parameters, objective_value, \
                            function_evaluations, statistics




"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2021 Marco Esposito, Leonardo Picchiami.
"""

from typing import List, Callable, Set

from apricopt.model.ObservableFunctionNotFoundError import ObservableFunctionNotFoundError


class ObservableFunction:
    @staticmethod
    def Identity(*args: List[float]) -> float:
        return ObservableFunction.LastValue(*args)

    @staticmethod
    def LastValue(*args: List[float]) -> float:
        return args[0][-1]

    @staticmethod
    def Average(*args: List[float]) -> float:
        if len(args[0]) == 0:
            return 0
        return sum(args[0]) / len(args[0])

    @staticmethod
    def Min(*args: List[float]) -> float:
        return min(args[0])

    @staticmethod
    def Max(*args: List[float]) -> float:
        return max(args[0])

    @staticmethod
    def Sum(*args: List[float]) -> float:
        return sum(args[0])

    @staticmethod
    def MaxDistanceFromLowerBound(*args: List[float]) -> float:
        value: List[float] = args[0]
        lower_bound: List[float] = args[1]

        return max(lower_bound[t] - value[t] for t in range(len(value)))

    @staticmethod
    def MaxDistanceFromUpperBound(*args: List[float]):
        value: List[float] = args[0]
        upper_bound: List[float] = args[1]

        return max(value[t] - upper_bound[t] for t in range(len(value)))

    @staticmethod
    def MaxDistanceFromBounds(*args: List[float]) -> float:
        value: List[float] = args[0]
        lower_bound: List[float] = args[1]
        upper_bound: List[float] = args[2]

        return max(
            max(lower_bound[t] - value[t], value[t] - upper_bound[t]) for t in range(len(value)))

    @staticmethod
    def FinalDistanceFromUpperBound(*args: List[float]) -> float:
        value: float = args[0][-1]
        upper_bound: float = args[1][-1]
        return value - upper_bound

    @staticmethod
    def CumSumSpikes(*args: List[float]):
        value: List[float] = args[0]

        cumsum: float = value[0]
        for t in range(1, len(value)):
            if value[t] > value[t - 1]:
                cumsum += value[t] - value[t - 1]
        return cumsum

    @staticmethod
    def CumSumSpikesDistanceFromLowerBound(*args: List[float]) -> float:
        cumsum: float = ObservableFunction.CumSumSpikes(*args)
        return args[1][-1] - cumsum

    @staticmethod
    def CumSumSpikesDistanceFromUpperBound(*args: List[float]) -> float:
        cumsum: float = ObservableFunction.CumSumSpikes(*args)
        return cumsum - args[1][-1]

    @staticmethod
    def CumSumSpikesDistanceFromBounds(*args: List[float]) -> float:
        cumsum: float = ObservableFunction.CumSumSpikes(*args)
        return max(args[1][-1] - cumsum, cumsum - args[2][-1])

    @staticmethod
    def MaxDifferenceInPeriod(*args: List[float]) -> float:
        value: List[float] = args[0]
        period_length: int = int(args[1][0])

        return max(value[t] - value[t - period_length - 1] for t in range(period_length, len(value)))

    @staticmethod
    def MaxDifferenceInPeriodDistanceFromAbove(*args: List[float]) -> float:
        return ObservableFunction.MaxDifferenceInPeriod(*args) - args[2][0]

    @staticmethod
    def MaxSpikesSumDifferenceInPeriodDistanceFromAbove(*args: List[float]) -> float:
        value: List[float] = args[0]
        period_length = args[1]
        upper_bound = args[2]

        cumsum: List[float] = [value[0]]
        for t in range(1, len(value)):
            if value[t] > value[t - 1]:
                cumsum[t] = value[t] - value[t - 1]
            else:
                cumsum[t] = value[t - 1]
        return ObservableFunction.MaxDifferenceInPeriodDistanceFromAbove(cumsum, period_length, upper_bound)

    @staticmethod
    def IndexOfFirstGE(*args: List[float]) -> float:
        value: List[float] = args[0]
        lower_bound: List[float] = args[1]
        for t in range(len(value)):
            if value[t] >= lower_bound[t]:
                return t
        return len(value)

    @staticmethod
    def IndexOfFirstGT(*args: List[float]) -> float:
        value: List[float] = args[0]
        lower_bound: List[float] = args[1]
        for t in range(len(value)):
            if value[t] > lower_bound[t]:
                return t
        return len(value)

    @staticmethod
    def IndexOfFirstEQ(*args: List[float]) -> float:
        value: List[float] = args[0]
        target: List[float] = args[1]
        for t in range(len(value)):
            if value[t] == target[t]:
                return t
        return len(value)

    @staticmethod
    def IndexOfFirstLE(*args: List[float]) -> float:
        value: List[float] = args[0]
        upper_bound: List[float] = args[1]
        for t in range(len(value)):
            if value[t] <= upper_bound[t]:
                return t
        return len(value)

    @staticmethod
    def IndexOfFirstLT(*args: List[float]) -> float:
        value: List[float] = args[0]
        upper_bound: List[float] = args[1]
        for t in range(len(value)):
            if value[t] < upper_bound[t]:
                return t
        return len(value)

    __function_names = {"max": Max,
                        "min": Min,
                        "avg": Average, "mean": Average,
                        "maxdistancefrombounds": MaxDistanceFromBounds,
                        "maxdistancefromlowerbound": MaxDistanceFromLowerBound,
                        "maxdistancefromupperbound": MaxDistanceFromUpperBound,
                        "cumsumspikes": CumSumSpikes,
                        "maxdifferenceinperiod": MaxDifferenceInPeriod,
                        "maxdifferenceinperioddistancefromabove": MaxDifferenceInPeriodDistanceFromAbove,
                        "maxspikessumdifferenceinperioddistancefromabove":
                            MaxSpikesSumDifferenceInPeriodDistanceFromAbove,
                        "indexoffirstge": IndexOfFirstGE,
                        "indexoffirstgt": IndexOfFirstGT,
                        "indexoffirsteq": IndexOfFirstEQ,
                        "indexoffirstle": IndexOfFirstLE,
                        "indexoffirstlt": IndexOfFirstLT,
                        "finaldistancefromupperbound": FinalDistanceFromUpperBound}

    @staticmethod
    def get_function(function_name: str) -> Callable:
        if function_name.lower() in ObservableFunction.__function_names:
            return ObservableFunction.__function_names[function_name.lower()]
        else:
            raise ObservableFunctionNotFoundError(function_name)

    @staticmethod
    def add_function(function_name: str, function: Callable):
        if function_name not in ObservableFunction.__function_names:
            ObservableFunction.__function_names[function_name.lower()] = function

    @staticmethod
    def get_function_names() -> Set[str]:
        return {n for n in ObservableFunction.__function_names.keys()}



"""
This file is part of Apricopt.

Apricopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Apricopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Apricopt.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2020-2021 Marco Esposito, Leonardo Picchiami.
"""

import copy
from typing import Dict, Set, List
from .FastObservable import FastObservable
from .LogObservable import LogObservable
from .ModelInstance import ModelInstance
from .Observable import Observable, compute_expressions_values
from .Parameter import Parameter


class Model:
    """
    Objects of this class represents models of systems that can be simulated and that may be subject to
    simulation-based optimization.

    A Model object may have an objective function and one or more constraints.
    Constraints are subdivided into regular constraints and fast constraints.
    A regular constraint is a constraint whose value can only be computed via simulation of the modes. A fast constraint,
    on the other side, does not require a simulation to be computed.
    A Model object contains the definition of its parameter space, as a set of Parameter objects.
    An object of this class has a reference to a SimulationEngine object that actually simulates it and to a
    ModelInstance object that represents the actual object that is used to simulate the system.
    In some sense, the ModelObject provides a wrapping over the ModelInstance object, which may use other frameworks to
    perform simulations and other tasks.
    Finally, the object contains further information, i.e. the integration tolerances,
    the time step, the name of the model outputs that are actually observedand the name of the file that contains the
    model (if any).
    """

    def __init__(self, sim_engine, model_filename: str, abs_tol: float,
                 rel_tol: float, time_step: float, observed_outputs: List[str] = None,
                 observed_outputs_observable: List[Observable] = None, **exec_info):
        self.sim_engine = sim_engine
        self.model_filename: str = model_filename
        if sim_engine:
            self.instance: ModelInstance = sim_engine.load_model(model_filename, **exec_info)
            self.instance.set_simulation_configuration(abs_tol, rel_tol, time_step)
            self.initial_values = self.build_initial_values()
        self.absolute_tolerance = abs_tol
        self.relative_tolerance = rel_tol
        self.time_step = time_step
        self.parameters: Dict[str, Parameter] = dict()
        self.objective: Observable = None
        self.constraints: Set[Observable] = set()
        self.fast_constraints: Set[FastObservable] = set()
        self.log_observables: Set[LogObservable] = set()
        self.cached_ids = None
        self.observed_outputs: List[str] = list() if observed_outputs is None else observed_outputs
        self.observed_outputs_observable: List[Observable] = set() if observed_outputs_observable is None \
                                                                                    else observed_outputs_observable

    def reinit(self):
        """
        Re-initializes the state of the object of type ModelInstance in the <pre>instance</pre> field.
        :return:
        """
        self.instance.__delete__(self)
        self.instance = self.sim_engine.load_model(self.model_filename)
        self.instance.set_simulation_configuration(self.absolute_tolerance, self.relative_tolerance, self.time_step)
        self.initial_values = self.build_initial_values()

    def _new_model_instance(self) -> None:
        """
        Initializes the <pre>instancce</pre> field with a ModelInstance object built starting from the model_filename.
        :return: None
        """
        self.instance = self.sim_engine.load_model(self.model_filename)
        self.instance.set_simulation_configuration(self.absolute_tolerance, self.relative_tolerance, self.time_step)
        self.initial_values = self.build_initial_values()

    def copy(self):
        """
        Performs a copy of self. This new object is a shallow copy of self, but the model instance of the copy is a
        new model instance of the same system.
        :return: None
        """
        copied_model = copy.copy(self)
        copied_model._new_model_instance()
        return copied_model

    def deep_copy(self):
        #TODO: returns a completely different Model object
        pass

    def set_parameter_space(self, params: Set[Parameter]) -> None:
        """
        Updates the parameter space using the Parameter objects in the given <pre>params</pre> set.
        :param params: a set of Parameter object. Each Parameter will be added to the parameter space of self.
        :return: None
        """
        for param in params:
            self.parameters[param.id] = param

    def set_params(self, params_values: Dict[str, float]) -> None:
        for param_id, value in params_values.items():
            param_obj = self.parameters[param_id]
            if param_obj.distribution == "uniform":
                if value < param_obj.lower_bound or \
                        value > param_obj.upper_bound:
                    raise ValueError(f"The value {value} for parameter {param_obj.name} is out of bounds.")
        self.instance.set_parameters(params_values)
        self.initial_values = self.build_initial_values()

    def set_parameters_nominal_values(self, params_values: Dict[str, float]) -> None:
        for param_id, value in params_values.items():
            param_obj = self.parameters[param_id]
            if param_obj.distribution == "uniform":
                if value < param_obj.lower_bound or \
                        value > param_obj.upper_bound:
                    raise ValueError(f"The value {value} for parameter {param_obj.name} is out of bounds.")
            param_obj.nominal_value = value

    def build_initial_values(self) -> Dict[str, float]:
        return dict(self.instance.get_parameters_initial_values(),
                    **self.instance.get_compartment_initial_volumes())

    def set_fixed_params(self, params_values: Dict[str, float]) -> None:
        # for param_id, value in params_values.items(): # commented because inefficient: one single set of all parameters is much quicker than many set of one parameter
        # self.instance.set_parameter(param_id, value)
        # self.initial_values[param_id] = value         WRONG: does not consider inter-parameter dependencies
        self.instance.set_parameters(params_values)
        self.initial_values = self.build_initial_values()

    def evaluate_fast_constraints(self, params_values: Dict[str, float]) -> Dict[str, float]:
        result: Dict[str, float] = dict()
        full_assignment = self.complete_assignment(params_values)
        for fast_constraint in self.fast_constraints:
            result[fast_constraint.id] = fast_constraint.evaluate(full_assignment)
        return result

    def build_zero_sim_output(self) -> Dict[str, float]:
        sim_output: Dict[str, float] = dict()

        if self.objective:
            sim_output[self.objective.id] = self.objective.upper_bound

        for constraint in self.constraints:
            sim_output[constraint.id] = constraint.upper_bound
         
        ''''#: ToCheck    
        for feasibility_constraint in self.feasibility_constraints:
            sim_output[feasibility_constraint.id] = feasibility_constraint.upper_bound'''
        return sim_output

    def get_observables_ids(self) -> Set[str]:
        if not self.cached_ids:
            self.cached_ids: Set[str] = set()
            if self.objective:
                self.cached_ids.add(self.objective.id)

            for constraint in self.constraints:
                self.cached_ids.add(constraint.id)

        return self.cached_ids

    def evaluate_constraints(self, trajectory: Dict[str, List[float]]) -> Dict[str, float]:
        full_trajectory: Dict[str, List[float]] = self.complete_trajectory(trajectory)
        result: Dict[str, float] = dict()
        if self.objective:
            result[self.objective.id] = self.objective.evaluate(full_trajectory)
        for constraint in self.constraints:
            result[constraint.id] = constraint.evaluate(full_trajectory)
        
        '''# ToCheck
        for feasibility_constraint in self.feasibility_constraints:
            result[feasibility_constraint.id] = feasibility_constraint.evaluate(full_trajectory)
        '''
        return result

    def evaluate_log_observables(self, trajectory: Dict[str, List[float]]) -> Dict[str, str]:
        full_trajectory: Dict[str, List[float]] = self.complete_trajectory(trajectory)
        result: Dict[str, str] = dict()
        for log_observable in self.log_observables:
            result[log_observable.id] = log_observable.evaluate(full_trajectory)
        return result

    def complete_trajectory(self, trajectory: Dict[str, List[float]]) -> Dict[str, List[float]]:
        length = len(trajectory['time'])
        full_trajectory: Dict[str, List[float]] = dict(trajectory)
        for value_id, value in self.initial_values.items():
            if value_id not in full_trajectory:
                full_trajectory[value_id] = [value] * length

        return full_trajectory

    def complete_assignment(self, assignment: Dict[str, float]) -> Dict[str, float]:
        full_assignment: Dict[str, float] = dict(assignment)
        for value_id, value in self.initial_values.items():
            if value_id not in full_assignment:
                full_assignment[value_id] = value
        return full_assignment

    def compute_first_satisfaction_index(self, obs: Observable, trajectory: Dict[str, List[float]]) -> int:
        # assume that function is Identity, so only one expression
        if len(obs.expressions) > 1:
            raise ValueError("An initialization constraint cannot have more than one expression in its formula")

        full_trajectory: Dict[str, List[float]] = self.complete_trajectory(trajectory)
        constraint_values = compute_expressions_values(full_trajectory, obs.expressions, obs.function)[0]
        trajectory_length: int = len(full_trajectory['time'])

        for t in range(trajectory_length):
            if constraint_values[t] <= 0:
                return t
        return trajectory_length
